
/*--------------------------------------------------------------------*/
/*--- inst_fetch: The instruction trace generation tool. if_main.c ---*/
/*--------------------------------------------------------------------*/

/*
   This file is part of inst_fetch a very simple Valgrind tool,
   which generates a trace of instruction fetches.

   Copyright (C) 2002-2004 Nicholas Nethercote
      njn25@cam.ac.uk

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/

#include "vg_skin.h"
#include <stdlib.h>
#include <stdio.h>

static Char* trace_out_file;
static Int     fd;

static REGPARM(1)
void log_data(UChar* string)
{
   VG_(write)(fd, (void*)string, VG_(strlen)(string));
}

static void file_err ( void )
{
   VG_(message)(Vg_UserMsg,
      "error: can't open trace output file `%s'",
      trace_out_file );
   VG_(message)(Vg_UserMsg,
      "       ... so trace  will be generated.");
}

void SK_(pre_clo_init)(void)
{
   Char* base_dir = NULL;

   VG_(details_name)            ("inst_fetch");
   VG_(details_version)         (NULL);
   VG_(details_description)     ("an instruction trace generator");
   VG_(details_copyright_author)(
      "Copyright (C) 2002-2004, and GNU GPL'd, by William Cohen.");
   VG_(details_bug_reports_to)  (VG_BUGS_TO);

   VG_(register_compact_helper)((Addr) & log_data);

   /* Get working directory */
   sk_assert( VG_(getcwd_alloc)(&base_dir) );

   /* Block is big enough for dir name + trace.out.<pid> */
   trace_out_file = VG_(malloc)((VG_(strlen)(base_dir) + 32)*sizeof(Char));
   VG_(sprintf)(trace_out_file, "%s/trace.out.%d",
                base_dir, VG_(getpid)());
   VG_(free)(base_dir);

   fd = VG_(open)(trace_out_file, VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,
                                       VKI_S_IRUSR|VKI_S_IWUSR);

   if (fd < 0) {
      // If the file can't be opened for whatever reason (conflict
      // between multiple cachegrinded processes?), give up now.
      file_err(); 
      return;
   }

}

void SK_(post_clo_init)(void)
{
}

/*------------------------------------------------------------*/
/*--- Data Collection functions                            ---*/
/*------------------------------------------------------------*/

UCodeBlock* SK_(instrument)(UCodeBlock* cb_in, Addr orig_addr)
{
   UCodeBlock* cb;
   UInstr*     u_in;
   Int         i;
   Int         addr_digits = sizeof(Addr) * 2;
   UChar*      string;
   Int	       str_buffer_size;
   UChar       hexdigit[16]="0123456789abcdef";
   UInt        x86_instr_size;
   Int         str_pos=0;
   Bool        instrumented_Jcc = False;

   cb = VG_(setup_UCodeBlock)(cb_in);

   /* allocate space for the string */
   str_buffer_size = addr_digits + 1 + VG_(get_num_instrs)(cb_in) + 1 + 1;
   string = (UChar *) malloc (str_buffer_size);
   if (!string) {
	      VG_(message)(Vg_UserMsg,
			   "error: cannot allocate space for string");
	      exit(EXIT_FAILURE);
   }

   /* put in the starting address into string */
   switch (addr_digits){
   case 8:
	   sprintf(string, "%8.0x:\n", orig_addr);
	   break;
   case 16:
	   sprintf(string, "%16.0x:\n", orig_addr);
	   break;
   default:
	   VG_(message)(Vg_UserMsg,
			"error: cannot handle pointer with %d digits",
			addr_digits);
	   exit(EXIT_FAILURE);
   }
   str_pos = addr_digits + 1;

   for (i = 0; i < VG_(get_num_instrs)(cb_in); i++) {
      u_in = VG_(get_instr)(cb_in, i);
      switch (u_in->opcode) {
      case INCEIP:
	      /* record size of the instructions (1 to 15 bytes) */
	      x86_instr_size = u_in->val1;
	      goto count_x86_instr_size;
      case JMP:
            if (instrumented_Jcc) {
               sk_assert(CondAlways == u_in->cond);
               sk_assert(i+1 == VG_(get_num_instrs)(cb_in));
               instrumented_Jcc = False;     // rest
               break;
            } else {
               // The first JMP... instrument.
               if (CondAlways != u_in->cond) {
                  sk_assert(i+2 == VG_(get_num_instrs)(cb_in));
                  instrumented_Jcc = True;
               } else {
                  sk_assert(i+1 == VG_(get_num_instrs)(cb_in));
               }
               // Get x86 instr size from final JMP.
	      /* record size of the instructions (1 to 15 bytes) */
	       x86_instr_size = VG_(get_last_instr)(cb_in)->extra4b;
	      goto count_x86_instr_size;
            }
      count_x86_instr_size:
	      /* We could replace the following with a simulator function
		 to avoid writing out gazillion bytes to file.
	       */
	       string[str_pos++] = hexdigit[x86_instr_size];
	       break;
      default:
	      break;
      }

   }
   string[str_pos++]='\n';
   string[str_pos++]='\0';

   {
	   /* generate code to log string */
	   Int     t_CC_addr;

	   t_CC_addr = newTemp(cb);
	   uInstr2(cb, MOV,   4, Literal, 0, TempReg, t_CC_addr);
	   uLiteral(cb, (Addr)string);

	   uInstr1(cb, CCALL, 0, TempReg, t_CC_addr);
	   uCCall(cb, (Addr) &log_data, 1, 1, False);
   }

   /* copy over the rest of the ucode */
   for (i = 0; i < VG_(get_num_instrs)(cb_in); i++) {
      u_in = VG_(get_instr)(cb_in, i);
      VG_(copy_UInstr)(cb, u_in);
   }

   VG_(free_UCodeBlock)(cb_in);

   return cb;
}

void SK_(fini)(Int exitcode)
{
   VG_(close)(fd);
}

VG_DETERMINE_INTERFACE_VERSION(SK_(pre_clo_init), 0)

/*--------------------------------------------------------------------*/
/*--- end                                                          ---*/
/*--------------------------------------------------------------------*/
